from pydantic import BaseModel
import pytest
from syrupy.extensions.json import JSONSnapshotExtension

from commonfate_provider.runtime import AWSLambdaRuntime
from commonfate_provider import namespace, provider, access, target, resources, tasks
from commonfate_provider.tests import helper


@pytest.fixture
def snapshot_json(snapshot):
    """use JSON, rather than AmberSnapshotExtension as our schema is serialized as JSON"""
    return snapshot.use_extension(JSONSnapshotExtension)


@pytest.fixture(autouse=True)
def fresh_namespace():
    yield
    namespace.clear()


@pytest.fixture
def runtime_fixture():
    class BasicProvider(provider.Provider):
        pass

    @access.target(kind="Default")
    class Args:
        group = target.String()
        pass

    @access.grant(kind="Default")
    def grant(p: BasicProvider, subject: str, target: Args):
        pass

    basic_provider = BasicProvider()

    runtime = AWSLambdaRuntime(provider=basic_provider)
    return runtime


def test_lambda_handler_works(runtime_fixture: AWSLambdaRuntime):
    event = {
        "type": "grant",
        "data": {
            "subject": "testuser",
            "target": {"arguments": {"group": "test"}, "kind": "Default"},
        },
    }
    runtime_fixture.handle(event=event, context=None)


def test_provider_describe(runtime_fixture: AWSLambdaRuntime, snapshot_json):
    event = {"type": "describe"}
    actual = runtime_fixture.handle(event=event, context=None)
    assert actual == snapshot_json


def test_provider_describe_with_config(snapshot_json):
    class Provider(provider.Provider):
        api_url = provider.String(description="API URL")
        api_key = provider.String(description="API key", secret=True)

    p = helper.initialise_test_provider(
        {"api_url": "https://example.com", "api_key": "abcdef"}
    )
    runtime = AWSLambdaRuntime(
        provider=p,
        publisher="acmecorp",
        name="test",
        version="v0.1.0",
        schema_version="v1",
    )

    event = {"type": "describe"}
    actual = runtime.handle(event=event, context=None)
    assert actual == snapshot_json


def test_provider_describe_with_errors(snapshot_json):
    class Provider(provider.Provider):
        api_url = provider.String(description="API URL")
        api_key = provider.String(description="API key", secret=True)

    p = helper.initialise_test_provider(
        {"api_url": "https://example.com", "api_key": "abcdef"}
    )
    runtime = AWSLambdaRuntime(
        provider=p,
        publisher="acmecorp",
        name="test",
        version="v0.1.0",
        schema_version="v1",
    )
    p.diagnostics.error("some error happened!")

    event = {"type": "describe"}
    actual = runtime.handle(event=event, context=None)
    assert actual == snapshot_json


def test_provider_describe_with_resources(snapshot_json):
    class Provider(provider.Provider):
        pass

    class MyResource(resources.Resource):
        val: str

    class OtherResource(resources.BaseResource):
        other_val: str

    p = Provider()
    runtime = AWSLambdaRuntime(provider=p)

    event = {"type": "describe"}
    actual = runtime.handle(event=event, context=None)
    assert actual == snapshot_json


def test_load_works(snapshot_json):
    class Provider(provider.Provider):
        pass

    class MyResource(resources.Resource):
        val: str

    @resources.loader
    def example_loader(p: Provider):
        resources.register(MyResource(id="123", name="name", val="first"))
        resources.register(MyResource(id="456", val="second", name="resource name"))

    p = Provider()

    runtime = AWSLambdaRuntime(
        provider=p,
    )

    event = {"type": "load", "data": {"task": "example_loader"}}
    actual = runtime.handle(event=event, context=None)
    assert actual == snapshot_json


def test_load_works_with_subtasks(snapshot_json):
    class Provider(provider.Provider):
        pass

    class MyResource(resources.Resource):
        val: str

    class MyTask(tasks.Task):
        val: str

        def run(self, p: Provider):
            pass

    @resources.loader
    def example_loader(p: Provider):
        tasks.call(MyTask(val="test"))

    p = Provider()

    runtime = AWSLambdaRuntime(
        provider=p,
    )

    event = {"type": "load", "data": {"task": "example_loader"}}
    actual = runtime.handle(event=event, context=None)
    assert actual == snapshot_json


def test_grant_works(snapshot_json):
    class Provider(provider.Provider):
        pass

    @access.target()
    class Target:
        pass

    class State(BaseModel):
        something: str

    @access.grant()
    def grant(p: Provider, subject: str, target: Target) -> access.GrantResult:
        return access.GrantResult(
            access_instructions="test", state=State(something="something")
        )

    p = Provider()

    runtime = AWSLambdaRuntime(
        provider=p,
    )

    event = {
        "type": "grant",
        "data": {"subject": "testuser", "target": {"kind": "Target", "arguments": {}}},
    }
    actual = runtime.handle(event=event, context=None)
    assert actual == snapshot_json


def test_revoke_works_with_state():
    class Provider(provider.Provider):
        pass

    @access.target()
    class Target:
        pass

    class State(BaseModel):
        something: str

    got_var = None

    @access.revoke()
    def revoke(
        p: Provider, subject: str, target: Target, state: State
    ) -> access.GrantResult:
        nonlocal got_var
        got_var = state.something

    p = Provider()

    runtime = AWSLambdaRuntime(
        provider=p,
    )

    event = {
        "type": "revoke",
        "data": {
            "state": {"something": "test"},
            "subject": "testuser",
            "target": {"kind": "Target", "arguments": {}},
        },
    }
    actual = runtime.handle(event=event, context=None)

    # this should be set if the revoke function was
    # called with the correct state.
    assert got_var == "test"



import typing

import pytest
from syrupy.extensions.json import JSONSnapshotExtension

from commonfate_provider import namespace, resources


@pytest.fixture(autouse=True)
def fresh_namespace():
    """clear the registered provider, targets etc between test runs"""
    yield
    namespace.clear()


@pytest.fixture
def snapshot_json(snapshot):
    """use JSON, rather than AmberSnapshotExtension as our schema is serialized as JSON"""
    return snapshot.use_extension(JSONSnapshotExtension)


def test_resource_schema_works(snapshot_json):
    class FirstResource(resources.Resource):
        """
        An example resource
        """

        value: str
        related_field: typing.Optional[str] = resources.Related("FirstResource")

    class SecondResource(resources.Resource):
        value: str
        related_field: typing.Optional[str] = resources.Related(FirstResource)

    @resources.loader
    def load_resources():
        pass

    got = resources.export_schema()
    assert got.dict() == snapshot_json


def test_resource_schema_works_with_unnamed_resources(snapshot_json):
    class NoName(resources.BaseResource):
        """
        An example resource
        """

        value: str

    got = resources.export_schema()
    assert got.dict() == snapshot_json



import typing
from commonfate_provider import namespace, tasks
from pydantic import BaseModel, Field
import inspect
from common_fate_schema.provider import v1alpha1


def composite_id(fields: typing.List[str]):
    return "/".join(fields)


class BaseResource(BaseModel):
    #  A resource with no name. Usually you'll want to subclass resources.Resource instead.

    id: str = Field(title="ID")

    def __init_subclass__(cls) -> None:
        namespace.register_resource_class(cls)
        return super().__init_subclass__()

    def export_json(self) -> dict:
        data = dict(self)
        id = data.pop("id")
        output = {"type": self.__class__.__name__, "id": id, "data": data}
        return output

    class Config:
        @staticmethod
        def schema_extra(
            schema: typing.Dict[str, typing.Any], model: typing.Type["BaseResource"]
        ) -> None:
            """
            shift all of the properties that aren't 'id' or 'name' into a 'data'
            field in the schema to match the export_json() method
            """

            data_keys = [
                key
                for key in schema.get("properties", {}).keys()
                if key != "id" and key != "name"
            ]

            if len(data_keys) == 0:
                return

            schema["properties"]["data"] = {}

            for key in data_keys:
                subschema = schema["properties"].pop(key)
                schema["properties"]["data"][key] = subschema


class Resource(BaseResource):
    name: str

    def export_json(self) -> dict:
        data = dict(self)
        id = data.pop("id")
        name = data.pop("name")
        output = {"type": self.__class__.__name__, "id": id, "name": name, "data": data}
        return output


T = typing.TypeVar("T", bound=BaseResource)


def Related(
    to: typing.Union[typing.Type[T], str], title: str = None, description: str = None
) -> str:
    if inspect.isclass(to):
        to = to.__name__
    return Field(relation=to, title=title, description=description)


def loader(func: tasks.LoaderFunc):
    namespace.register_resource_loader(func)
    return func


def export_schema() -> v1alpha1.Resources:
    """
    Returns the schema for the 'resources' section in a provider schema as a dict.
    This section defines the resources the provider can read, along with the
    fetching methods which can be called to fetch them.
    """
    loaders: typing.Dict[str, v1alpha1.Loader] = {}
    for k in namespace.get_resource_loaders().keys():
        loaders[k] = v1alpha1.Loader(title=k)

    resources = v1alpha1.Resources(loaders=loaders, types={})
    for Klass in namespace.get_resource_classes():
        properties = {}
        all_vars = [(k, v) for (k, v) in vars(Klass).items() if not k.startswith("__")]
        for k, v in all_vars:
            properties[k] = {"type": "string"}
        resources.types[Klass.__name__] = Klass.schema()

    return resources


def register(resource: BaseResource):
    namespace._ALL_RESOURCES.append(resource)


def get() -> typing.List[BaseResource]:
    return namespace._ALL_RESOURCES


def _reset():
    namespace._ALL_RESOURCES = []


def without_keys(d, keys):
    return {x: d[x] for x in d if x not in keys}



from dataclasses import dataclass
import inspect
import typing_extensions
import typing

from pydantic import BaseModel

from commonfate_provider import namespace, provider, rpc
from commonfate_provider import target as cf_target


_T = typing.TypeVar("_T")


@typing_extensions.dataclass_transform()
def target(
    kind: typing.Optional[str] = None,
) -> typing.Callable[[type[_T]], type[_T]]:
    """
    Define a target for access.

    Example:
    ```
    @access.target()
    class Target:
        target_property = target.String()
    ```
    """

    def actual_decorator(cls: type[_T]) -> type[_T]:
        nonlocal kind
        if kind is None:
            kind = cls.__name__

        namespace.register_target_class(kind, cls)

        return cls

    return actual_decorator


class GrantResult(BaseModel):
    access_instructions: typing.Optional[str] = None
    """
    Instructions on how to access the entitlements.
    Common Fate will display these to the user upon a successful Access Request.
    """
    state: typing.Optional[BaseModel] = None
    """
    State which will be stored and provided to the corresponding revoke function.
    """


GrantValidatorFunc = typing.Callable[[typing.Any, str, typing.Any], None]


@dataclass
class GrantValidator:
    name: str
    func: GrantValidatorFunc


def grant_validator(
    name: str,
    kind: namespace.KindType = None,
) -> typing.Callable[[GrantValidatorFunc], GrantValidatorFunc]:
    def actual_decorator(func: GrantValidatorFunc):
        id = func.__name__
        gv = GrantValidator(id=id, name=name, func=func)
        namespace.register_grant_validator(kind=kind, grant_validator=gv)
        return func

    return actual_decorator


GrantFunc = typing.Callable[[typing.Any, str, typing.Any], typing.Optional[GrantResult]]

_T = typing.TypeVar("_T")


def grant(
    kind: namespace.KindType = None,
) -> typing.Callable[[GrantFunc], GrantFunc]:
    """
    Register a function as an access granter.

    For example:
    ```
    @access.grant()
    def grant(p: Provider, subject: str, target: MyTargetClass):
        ...
    ```

    The `kind` parameter may be specified to indicate a particular target kind
    that this function grants access to.

    For example:
    ```
    @access.target(kind="MyTargetKind")
    class Target:
        ...

    @access.grant(kind="MyTargetKind")
    def grant(p: Provider, subject: str, target: MyTargetClass):
        ...
    ```
    """

    def actual_decorator(func: GrantFunc):
        namespace.register_grant_func(kind=kind, func=func)
        return func

    return actual_decorator


RevokeFunc = typing.Callable[[typing.Any, str, typing.Any], None]


def revoke(
    kind: typing.Optional[typing.Union[typing.Type[_T], str]] = None,
) -> typing.Callable[[RevokeFunc], RevokeFunc]:
    """
    Register a function as an access revoker.

    For example:
    ```
    @access.revoke()
    def revoke(p: Provider, subject: str, target: MyTargetClass):
        ...
    ```

    The `kind` parameter may be specified to indicate a particular target kind
    that this function revokes access to.

    For example:
    ```
    @access.target(kind="MyTargetKind")
    class Target:
        ...

    @access.revoke(kind="MyTargetKind")
    def revoke(p: Provider, subject: str, target: MyTargetClass):
        ...
    ```
    """

    def actual_decorator(func: RevokeFunc):
        namespace.register_revoke_func(kind=kind, func=func)
        return func

    return actual_decorator


def call_access_func(
    type: typing.Union[typing.Literal["grant"], typing.Literal["revoke"]],
    p: provider.Provider,
    data: rpc.GrantData,
):
    """
    Calls the actual @access.grant() or @access.revoke function.

    The PDK supports optional 'state', and "request" which is passed between grant() and revoke().

    For a stateful revoke function, the signature looks like:
    ```
    def revoke(p: Provider, subject: str, target: Target, state: State):
        ...
    ```

    For a stateless revoke function, the signature looks like:
    ```
    def revoke(p: Provider, subject: str, target: Target):
        ...
    ```

    This method inspects the signature of the revoke function, and passes
    state through if the signature supports it.
    """
    try:
        registered_targets = namespace.get_target_classes()
        registered_target = registered_targets[data.target.kind]
        args_cls = registered_target.cls
    except KeyError:
        all_keys = ",".join(registered_targets.keys())
        raise KeyError(
            f"unhandled target kind {data.target.kind}, supported kinds are [{all_keys}]"
        )

    t = cf_target._initialise(args_cls, data.target.arguments)

    if type == "grant":
        func = registered_target.get_grant_func()
    else:
        func = registered_target.get_revoke_func()

    # initialise the arguments that the revoke() function
    # will be called by
    kwargs = {
        "p": p,
        "subject": data.subject,
        "target": t,
        "state": data.state,
        "request": data.request,
    }

    spec = inspect.getfullargspec(func)

    # check if the function uses state
    state_anno = spec.annotations.get("state", None)

    # if it doesn't, remove it from the arguments
    # the function is called with
    if state_anno is None:
        del kwargs["state"]

    elif issubclass(state_anno, BaseModel):
        kwargs["state"] = state_anno.parse_obj(data.state)

    # check if the function uses the request
    request_anno = spec.annotations.get("request", None)

    # if it doesn't, remove it from the arguments
    # the function is called with
    if request_anno is None:
        del kwargs["request"]

    return func(**kwargs)
